from __future__ import (
    absolute_import,
    unicode_literals,
)

from pymetrics.version import (
    __version__,
    __version_info__,
)


__all__ = (
    '__version__',
    '__version_info__',
)
